"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateCode = exports.printAst = exports.parseAst = exports.fileToAst = exports.makeSetState = exports.throwError = void 0;
const fs_1 = require("fs");
const parser_1 = require("@babel/parser");
const t = __importStar(require("@babel/types"));
const traverse_1 = __importDefault(require("@babel/traverse"));
const generator_1 = __importDefault(require("@babel/generator"));
const recurse_1 = require("./recurse");
const throwError = (str) => {
    throw new Error(`Genomic Parser Error: ${str}`);
};
exports.throwError = throwError;
const makeSetState = (name) => {
    return `set${name.charAt(0).toUpperCase() + name.slice(1)}`;
};
exports.makeSetState = makeSetState;
const fileToAst = (path, parserOptions = {
    sourceType: 'module',
    plugins: [
        'jsx', 'typescript'
    ]
}) => {
    return (0, parser_1.parse)((0, fs_1.readFileSync)(path, 'utf-8'), parserOptions);
};
exports.fileToAst = fileToAst;
const parseAst = (ast, clean = false) => {
    const result = clean ? (0, recurse_1.cleanAst)(ast) : ast;
    let configOptions = {};
    const componentHash = {};
    let componentCursor = null;
    // validations and loading state definitions
    (0, traverse_1.default)(result, {
        ImportDeclaration(path) {
            if (path.node.source.value === 'react') {
                (0, exports.throwError)('cannot import react!');
            }
        }
    });
    // get options
    (0, traverse_1.default)(result, {
        ExpressionStatement(path) {
            const expr = path.node.expression;
            if (expr.type === 'CallExpression' &&
                expr.callee.type === 'Identifier' &&
                expr.callee.name === 'useConfig') {
                if (path.parent.type !== 'Program') {
                    (0, exports.throwError)('bad config placement');
                }
                if (expr.arguments.length === 0) {
                    (0, exports.throwError)('no config args');
                }
                if (expr.arguments.length !== 1) {
                    (0, exports.throwError)('too many config args');
                }
                if (expr.arguments[0].type !== 'ObjectExpression') {
                    (0, exports.throwError)('config must be an object');
                }
                try {
                    // eval isn't always evil ;) compile time!
                    eval(`configOptions = ${(0, generator_1.default)(expr.arguments[0]).code}`);
                }
                catch (e) {
                    throw new Error('Genonomic Parser Error: bad config options');
                }
                path.remove();
                path.stop();
            }
        }
    });
    console.log({ configOptions });
    // now traverse :) 
    (0, traverse_1.default)(result, {
        ImportDeclaration(path) {
            /// useRef, useStore, useState, useEffect, userDefaultProps
            if (path.node.source.value === "@genomoic/stubs") {
                [
                    'useStore',
                    'useState',
                    'useRef',
                    'useDefaultProps',
                    'onMount',
                    'onUnmount',
                    'onUpdate'
                ].forEach(name => {
                    if (path.node.specifiers.find(a => a.type === 'ImportSpecifier' &&
                        a.imported.type === 'Identifier' &&
                        a.imported.name === name)) {
                        path.node.source.value = "react";
                    }
                });
            }
        },
        VariableDeclarator(path) {
            // manage cursor
            if (path.node.init &&
                path.node.init.type === 'CallExpression' &&
                path.node.init.callee &&
                path.node.init.callee.type === 'Identifier' &&
                (path.node.init.callee.name === 'useStore' ||
                    path.node.init.callee.name === 'useState' ||
                    path.node.init.callee.name === 'useRef')) {
                if (path.parent.type === 'VariableDeclaration') {
                    if (path.parentPath.parent.type === 'BlockStatement') {
                        const parent = path.parentPath.parentPath.parent;
                        const parentPath = path.parentPath.parentPath.parentPath;
                        if (parent.type === 'FunctionDeclaration') {
                            const componentName = parent.id.name;
                            componentCursor = componentName;
                            componentHash[componentName] = componentHash[componentName] || {
                                name: componentName,
                                refs: {},
                                state: {}
                            };
                        }
                        else if (parent.type === 'ArrowFunctionExpression') {
                            if (parentPath.parent.type === 'VariableDeclarator') {
                                if (parentPath.parent.id.type === 'Identifier') {
                                    const componentName = parentPath.parent.id.name;
                                    componentCursor = componentName;
                                    componentHash[componentName] = componentHash[componentName] || {
                                        name: componentName,
                                        refs: {},
                                        state: {}
                                    };
                                }
                                else {
                                    (0, exports.throwError)('cannot find ArrowFunctionExpression Component definition');
                                }
                            }
                        }
                        else {
                            (0, exports.throwError)('cannot find Component definition');
                        }
                    }
                }
            }
            // useRef transformer
            if (path.node.init &&
                path.node.init.type === 'CallExpression' &&
                path.node.init.callee &&
                path.node.init.callee.type === 'Identifier' &&
                path.node.init.callee.name === 'useRef') {
                const node = path.node;
                if (node.id.type !== 'Identifier') {
                    return (0, exports.throwError)('useRef must use an Identifier');
                }
                if (!componentCursor) {
                    console.log((0, generator_1.default)(node).code);
                }
                if (componentHash[componentCursor].refs[node.id.name] &&
                    componentHash[componentCursor].refs[node.id.name].hasOwnProperty('type')) {
                    // console.log(node.id)
                    // throwError('refs variable is already defined!')
                }
                else {
                    // TODO fix if case above so it can throw
                    componentHash[componentCursor].refs[node.id.name] = {
                        orig: 'useRef',
                        type: 'ObjectProperty',
                        used: false
                    };
                }
            }
            // useState transformer
            if (path.node.init &&
                path.node.init.type === 'CallExpression' &&
                path.node.init.callee &&
                path.node.init.callee.type === 'Identifier' &&
                path.node.init.callee.name === 'useState') {
                const node = path.node;
                if (node.id.type !== 'ArrayPattern') {
                    return (0, exports.throwError)('useState must use an array');
                }
                if (node.id.elements.length !== 2) {
                    return (0, exports.throwError)('useState must use an array of length 2');
                }
                if (node.id.elements[0].type !== 'Identifier') {
                    return (0, exports.throwError)('useState first element must be Identifier');
                }
                if (node.id.elements[1].type !== 'Identifier') {
                    return (0, exports.throwError)('useState second element must be Identifier');
                }
                if (!componentCursor) {
                    console.log((0, generator_1.default)(node).code);
                }
                if (componentHash[componentCursor].state[node.id.elements[0].name] &&
                    componentHash[componentCursor].state[node.id.elements[0].name].hasOwnProperty('type')) {
                    // console.log(node.id)
                    // throwError('state variable is already defined!')
                }
                else {
                    // TODO fix if case above so it can throw
                    componentHash[componentCursor].state[node.id.elements[0].name] = {
                        orig: 'useState',
                        type: 'ObjectProperty',
                        used: false
                    };
                }
            }
            // useStore transformer
            if (path.node.init &&
                path.node.init.type === 'CallExpression' &&
                path.node.init.callee &&
                path.node.init.callee.type === 'Identifier' &&
                path.node.init.callee.name === 'useStore') {
                // @ts-ignore
                const args = path.node.init.arguments;
                const objectArgument = args[0];
                if (t.isObjectExpression(objectArgument) && objectArgument.properties.length > 0) {
                    const newDeclarations = objectArgument.properties.map(property => {
                        switch (property.type) {
                            case 'ObjectMethod': {
                                // @ts-ignore
                                const method = property;
                                // @ts-ignore
                                const key = method.key;
                                if (key.type !== 'Identifier') {
                                    (0, exports.throwError)('store methods must be Identifiers');
                                }
                                if (componentHash[componentCursor].state[key.name] &&
                                    componentHash[componentCursor].state[key.name].hasOwnProperty('type')) {
                                    (0, exports.throwError)('state variable is already defined!');
                                }
                                componentHash[componentCursor].state[key.name] = {
                                    type: 'ObjectMethod',
                                    used: false
                                };
                                return t.functionDeclaration(t.identifier(key.name), method.params, method.body, method.generator, method.async);
                            }
                            case 'ObjectProperty': {
                                // @ts-ignore
                                const prop = property;
                                // @ts-ignore
                                const key = prop.key;
                                if (key.type !== 'Identifier') {
                                    (0, exports.throwError)('store object properties must be Identifier');
                                }
                                if (componentHash[componentCursor].state[key.name] &&
                                    componentHash[componentCursor].state[key.name].hasOwnProperty('type')) {
                                    (0, exports.throwError)('state variable is already defined!');
                                }
                                componentHash[componentCursor].state[key.name] = {
                                    type: 'ObjectProperty',
                                    used: false
                                };
                                const stateName = t.identifier(key.name);
                                const stateSetterName = t.identifier(`set${key.name.charAt(0).toUpperCase() + key.name.slice(1)}`);
                                return t.variableDeclaration('const', [
                                    t.variableDeclarator(t.arrayPattern([stateName, stateSetterName]), t.callExpression(t.identifier('useState'), [t.arrowFunctionExpression([], prop.value)]))
                                ]);
                            }
                            default:
                                (0, exports.throwError)('store must be ObjectMethod or ObjectProperty');
                        }
                        // @ts-ignore
                        const prop = property;
                        // @ts-ignore
                        const key = prop.key;
                        if (key.type !== 'Identifier') {
                            (0, exports.throwError)('store object properties must be Identifier');
                        }
                        const stateName = t.identifier(key.name);
                        const stateSetterName = t.identifier((0, exports.makeSetState)(key.name));
                        return t.variableDeclaration('const', [
                            t.variableDeclarator(t.arrayPattern([stateName, stateSetterName]), t.callExpression(t.identifier('useState'), [t.arrowFunctionExpression([], prop.value)]))
                        ]);
                    });
                    // this updates the array above! awesome!
                    path.parentPath.replaceWithMultiple(newDeclarations);
                }
            }
        },
        AssignmentExpression(path) {
            // STATE
            // FROM state.property = value;
            // TO   setProperty(value);
            if (path.node.type === 'AssignmentExpression') {
                if (path.node.left.type === 'MemberExpression') {
                    if (path.node.left.object.type === 'Identifier') {
                        if (path.node.left.object.name === 'state') {
                            // left: path.node.left.property as any
                            if (path.node.left.property.type !== 'Identifier') {
                                return (0, exports.throwError)('set state property not yet supported.');
                            }
                            // set that state variable is used!
                            if (componentHash[componentCursor].state[path.node.left.property.name] &&
                                componentHash[componentCursor].state[path.node.left.property.name].hasOwnProperty('used')) {
                                componentHash[componentCursor].state[path.node.left.property.name].used = true;
                            }
                            else {
                                (0, exports.throwError)('either component undefined or state set without being defined');
                            }
                            path.replaceWith(t.expressionStatement(t.callExpression(t.identifier((0, exports.makeSetState)(path.node.left.property.name)), [path.node.right])));
                        }
                    }
                }
            }
        },
        CallExpression(path) {
            // STATE
            // FROM state.method(value)
            // TO   method(value);
            if (path.node.type === 'CallExpression') {
                if (path.node.callee.type === 'MemberExpression') {
                    if (path.node.callee.object.type === 'Identifier') {
                        if (path.node.callee.object.name === 'state') {
                            // left: path.node.callee.property as any
                            if (path.node.callee.property.type !== 'Identifier') {
                                return (0, exports.throwError)('set state property not yet supported.');
                            }
                            // set that state variable is used!
                            if (componentHash[componentCursor].state[path.node.callee.property.name] &&
                                componentHash[componentCursor].state[path.node.callee.property.name].hasOwnProperty('used')) {
                                componentHash[componentCursor].state[path.node.callee.property.name].used = true;
                            }
                            else {
                                (0, exports.throwError)('either component undefined or state set without being defined');
                            }
                            path.replaceWith(t.callExpression(path.node.callee.property, path.node.arguments));
                        }
                    }
                }
            }
            // useDefault props transformer
            if (path.node.callee.type === 'Identifier' &&
                path.node.callee.name === 'useDefaultProps') {
                const args = path.node.arguments;
                // @ts-ignore
                const Component = args[0];
                // @ts-ignore
                const Props = args[1];
                if (Component.type !== 'Identifier') {
                    (0, exports.throwError)('useDefaultProps first argument must be an Identifier');
                }
                if (Props.type !== 'ObjectExpression') {
                    (0, exports.throwError)('useDefaultProps first argument must be an ObjectExpression');
                }
                // replace with Component.defaultProps = Props;
                path.replaceWith(t.expressionStatement(t.assignmentExpression('=', t.memberExpression(t.identifier(Component.name), t.identifier('defaultProps'), false), Props)));
            }
        },
        ExpressionStatement(path) {
            // FROM
            // onMount(() => {
            //     if (parentRef) {
            //         animateLayout(parentRef);
            //     }
            // });
            // TO 
            // useEffect(() => {
            //     if (parentRef.current) {
            //       animateLayout(parentRef.current);
            //     }
            //   }, []);
            if (path.node.expression.type === 'CallExpression') {
                if (path.node.expression.callee.type === 'Identifier') {
                    if (path.node.expression.callee.name === 'onMount') {
                        path.node.expression.callee.name = 'useEffect';
                        path.node.expression.arguments.push(t.arrayExpression([]));
                    }
                    if (path.node.expression.callee.name === 'onUnmount') {
                        path.node.expression.callee.name = 'useEffect';
                    }
                    if (path.node.expression.callee.name === 'onUpdate') {
                        path.node.expression.callee.name = 'useEffect';
                    }
                }
            }
        }
    });
    // update all the state!
    (0, traverse_1.default)(result, {
        MemberExpression(path) {
            if (path.node.object.type === 'Identifier' &&
                path.node.object.name === 'state') {
                path.replaceWith(path.node.property);
            }
        }
    });
    // update all the refs!
    (0, traverse_1.default)(result, {
        VariableDeclarator(path) {
            // manage cursor
            // MARKED AS NOT DRY
            if (path.node.init &&
                path.node.init.type === 'CallExpression' &&
                path.node.init.callee &&
                path.node.init.callee.type === 'Identifier' &&
                (path.node.init.callee.name === 'useStore' ||
                    path.node.init.callee.name === 'useState' ||
                    path.node.init.callee.name === 'useRef')) {
                if (path.parent.type === 'VariableDeclaration') {
                    if (path.parentPath.parent.type === 'BlockStatement') {
                        const parent = path.parentPath.parentPath.parent;
                        const parentPath = path.parentPath.parentPath.parentPath;
                        if (parent.type === 'FunctionDeclaration') {
                            const componentName = parent.id.name;
                            componentCursor = componentName;
                        }
                        else if (parent.type === 'ArrowFunctionExpression') {
                            if (parentPath.parent.type === 'VariableDeclarator') {
                                if (parentPath.parent.id.type === 'Identifier') {
                                    const componentName = parentPath.parent.id.name;
                                    componentCursor = componentName;
                                }
                                else {
                                    (0, exports.throwError)('cannot find ArrowFunctionExpression Component definition');
                                }
                            }
                        }
                        else {
                            (0, exports.throwError)('cannot find Component definition');
                        }
                    }
                }
            }
        },
        Identifier(path) {
            // in a MemberExpression?
            if (path.parent.type === 'MemberExpression') {
                return;
            }
            // in a VariableDeclarator?
            if (path.parent.type === 'VariableDeclarator') {
                return;
            }
            // is a ref the same name?
            if (!componentHash[componentCursor].refs.hasOwnProperty(path.node.name)) {
                return;
            }
            path.replaceWith(t.memberExpression(t.identifier(path.node.name), t.identifier('current')));
        }
    });
    return {
        store: componentHash,
        ast: result
    };
};
exports.parseAst = parseAst;
const printAst = (ast) => {
    const results = (0, exports.generateCode)(ast);
    console.log(results);
};
exports.printAst = printAst;
const generateCode = (ast) => {
    const results = (0, generator_1.default)(ast);
    return results.code;
};
exports.generateCode = generateCode;
