import traverse from "@babel/traverse";
import generate from '@babel/generator';
import { cleanAst } from './recurse';
import { defaultProps, lifecycleHooks, manageCursor, stateAssignments, stateFunctions, stubsToReact, updateRefsIdentifier, useRefVariableDeclarator, useStateVariableDeclarator, useStoreVariableDeclarator } from './transformers';
import { throwError } from './utils';
export const parseAst = (ast, clean = false) => {
    const result = clean ? cleanAst(ast) : ast;
    let configOptions = {};
    const componentHash = {};
    let componentCursor = null;
    // validations and loading state definitions
    traverse(result, {
        ImportDeclaration(path) {
            if (path.node.source.value === 'react') {
                throwError('cannot import react!');
            }
        }
    });
    // get options
    traverse(result, {
        ExpressionStatement(path) {
            const expr = path.node.expression;
            if (expr.type === 'CallExpression' &&
                expr.callee.type === 'Identifier' &&
                expr.callee.name === 'useConfig') {
                if (path.parent.type !== 'Program') {
                    throwError('bad config placement');
                }
                if (expr.arguments.length === 0) {
                    throwError('no config args');
                }
                if (expr.arguments.length !== 1) {
                    throwError('too many config args');
                }
                if (expr.arguments[0].type !== 'ObjectExpression') {
                    throwError('config must be an object');
                }
                try {
                    // eval isn't always evil ;) compile time!
                    eval(`configOptions = ${generate(expr.arguments[0]).code}`);
                }
                catch (e) {
                    throw new Error('Genonomic Parser Error: bad config options');
                }
                path.remove();
                path.stop();
            }
        }
    });
    console.log({ configOptions });
    // now traverse :) 
    traverse(result, {
        ImportDeclaration(path) {
            // useRef, useStore, useState, useEffect, userDefaultProps
            stubsToReact(path);
        },
        VariableDeclarator(path) {
            // update cursor
            componentCursor = manageCursor(path, {
                componentHash
            });
            // useRef transformer
            useRefVariableDeclarator(path, {
                componentCursor,
                componentHash
            });
            // useState transformer
            useStateVariableDeclarator(path, {
                componentCursor,
                componentHash
            });
            // useStore transformer
            useStoreVariableDeclarator(path, {
                componentCursor,
                componentHash
            });
        },
        AssignmentExpression(path) {
            // STATE
            // FROM state.property = value;
            // TO   setProperty(value);
            stateAssignments(path, {
                componentCursor,
                componentHash
            });
        },
        CallExpression(path) {
            // state.hide(el);
            stateFunctions(path, {
                componentCursor,
                componentHash
            });
            // useDefault props transformer
            defaultProps(path);
        },
        ExpressionStatement(path) {
            // onMount, onUnmount, onUpdate
            lifecycleHooks(path);
        }
    });
    // update all the state!
    traverse(result, {
        MemberExpression(path) {
            if (path.node.object.type === 'Identifier' &&
                path.node.object.name === 'state') {
                path.replaceWith(path.node.property);
            }
        }
    });
    // update all the refs!
    traverse(result, {
        VariableDeclarator(path) {
            // manage cursor
            componentCursor = manageCursor(path, {
                componentHash
            });
        },
        Identifier(path) {
            updateRefsIdentifier(path, {
                componentCursor,
                componentHash
            });
        }
    });
    return {
        store: componentHash,
        ast: result
    };
};
