import generate from '@babel/generator';
import * as t from '@babel/types';
import { makeSetState, throwError } from './utils';
export const stubsToReact = (path) => {
    /// useRef, useStore, useState, useEffect, userDefaultProps
    if (path.node.source.value === "@genomoic/stubs") {
        ['useRef', 'useStore', 'useState', 'useEffect', 'useDefaultProps'].forEach(name => {
            if (path.node.specifiers.find(a => a.type === 'ImportSpecifier' &&
                a.imported.type === 'Identifier' &&
                a.imported.name === name)) {
                path.node.source.value = "react";
            }
        });
    }
};
export const getConfigOptions = (path) => {
    const expr = path.node.expression;
    if (expr.type === 'CallExpression' &&
        expr.callee.type === 'Identifier' &&
        expr.callee.name === 'useConfig') {
        if (path.parent.type !== 'Program') {
            throwError('bad config placement');
        }
        if (expr.arguments.length === 0) {
            throwError('no config args');
        }
        if (expr.arguments.length !== 1) {
            throwError('too many config args');
        }
        if (expr.arguments[0].type !== 'ObjectExpression') {
            throwError('config must be an object');
        }
        try {
            // eval isn't always evil ;) compile time!
            eval(`configOptions = ${generate(expr.arguments[0]).code}`);
        }
        catch (e) {
            throw new Error('Genonomic Parser Error: bad config options');
        }
        path.remove();
        path.stop();
    }
    ;
    export const updateRefsIdentifier = (path, { componentCursor, componentHash }) => {
        // in a MemberExpression?
        if (path.parent.type === 'MemberExpression') {
            return;
        }
        // in a VariableDeclarator?
        if (path.parent.type === 'VariableDeclarator') {
            return;
        }
        // is a ref the same name?
        if (!componentHash[componentCursor].refs.hasOwnProperty(path.node.name)) {
            return;
        }
        path.replaceWith(t.memberExpression(t.identifier(path.node.name), t.identifier('current')));
    };
    export const useRefVariableDeclarator = (path, { componentCursor, componentHash }) => {
        if (path.node.init &&
            path.node.init.type === 'CallExpression' &&
            path.node.init.callee &&
            path.node.init.callee.type === 'Identifier' &&
            path.node.init.callee.name === 'useRef') {
            const node = path.node;
            if (node.id.type !== 'Identifier') {
                return throwError('useRef must use an Identifier');
            }
            if (componentHash[componentCursor].refs[node.id.name] &&
                componentHash[componentCursor].refs[node.id.name].hasOwnProperty('type')) {
                // console.log(node.id)
                // throwError('refs variable is already defined!')
            }
            else {
                // TODO fix if case above so it can throw
                componentHash[componentCursor].refs[node.id.name] = {
                    orig: 'useRef',
                    type: 'ObjectProperty',
                    used: false
                };
            }
        }
    };
    export const useStateVariableDeclarator = (path, { componentCursor, componentHash }) => {
        if (path.node.init &&
            path.node.init.type === 'CallExpression' &&
            path.node.init.callee &&
            path.node.init.callee.type === 'Identifier' &&
            path.node.init.callee.name === 'useState') {
            const node = path.node;
            if (node.id.type !== 'ArrayPattern') {
                return throwError('useState must use an array');
            }
            if (node.id.elements.length !== 2) {
                return throwError('useState must use an array of length 2');
            }
            if (node.id.elements[0].type !== 'Identifier') {
                return throwError('useState first element must be Identifier');
            }
            if (node.id.elements[1].type !== 'Identifier') {
                return throwError('useState second element must be Identifier');
            }
            if (!componentCursor) {
                console.log(generate(node).code);
            }
            if (componentHash[componentCursor].state[node.id.elements[0].name] &&
                componentHash[componentCursor].state[node.id.elements[0].name].hasOwnProperty('type')) {
                // console.log(node.id)
                // throwError('state variable is already defined!')
            }
            else {
                // TODO fix if case above so it can throw
                componentHash[componentCursor].state[node.id.elements[0].name] = {
                    orig: 'useState',
                    type: 'ObjectProperty',
                    used: false
                };
            }
        }
    };
    export const useStoreVariableDeclarator = (path, { componentCursor, componentHash }) => {
        if (path.node.init &&
            path.node.init.type === 'CallExpression' &&
            path.node.init.callee &&
            path.node.init.callee.type === 'Identifier' &&
            path.node.init.callee.name === 'useStore') {
            // @ts-ignore
            const args = path.node.init.arguments;
            const objectArgument = args[0];
            if (t.isObjectExpression(objectArgument) && objectArgument.properties.length > 0) {
                const newDeclarations = objectArgument.properties.map(property => {
                    switch (property.type) {
                        case 'ObjectMethod': {
                            // @ts-ignore
                            const method = property;
                            // @ts-ignore
                            const key = method.key;
                            if (key.type !== 'Identifier') {
                                throwError('store methods must be Identifiers');
                            }
                            if (componentHash[componentCursor].state[key.name] &&
                                componentHash[componentCursor].state[key.name].hasOwnProperty('type')) {
                                throwError('state variable is already defined!');
                            }
                            componentHash[componentCursor].state[key.name] = {
                                type: 'ObjectMethod',
                                used: false
                            };
                            return t.functionDeclaration(t.identifier(key.name), method.params, method.body, method.generator, method.async);
                        }
                        case 'ObjectProperty': {
                            // @ts-ignore
                            const prop = property;
                            // @ts-ignore
                            const key = prop.key;
                            if (key.type !== 'Identifier') {
                                throwError('store object properties must be Identifier');
                            }
                            if (componentHash[componentCursor].state[key.name] &&
                                componentHash[componentCursor].state[key.name].hasOwnProperty('type')) {
                                throwError('state variable is already defined!');
                            }
                            componentHash[componentCursor].state[key.name] = {
                                type: 'ObjectProperty',
                                used: false
                            };
                            const stateName = t.identifier(key.name);
                            const stateSetterName = t.identifier(`set${key.name.charAt(0).toUpperCase() + key.name.slice(1)}`);
                            return t.variableDeclaration('const', [
                                t.variableDeclarator(t.arrayPattern([stateName, stateSetterName]), t.callExpression(t.identifier('useState'), [t.arrowFunctionExpression([], prop.value)]))
                            ]);
                        }
                        default:
                            throwError('store must be ObjectMethod or ObjectProperty');
                    }
                    // @ts-ignore
                    const prop = property;
                    // @ts-ignore
                    const key = prop.key;
                    if (key.type !== 'Identifier') {
                        throwError('store object properties must be Identifier');
                    }
                    const stateName = t.identifier(key.name);
                    const stateSetterName = t.identifier(makeSetState(key.name));
                    return t.variableDeclaration('const', [
                        t.variableDeclarator(t.arrayPattern([stateName, stateSetterName]), t.callExpression(t.identifier('useState'), [t.arrowFunctionExpression([], prop.value)]))
                    ]);
                });
                // this updates the array above! awesome!
                path.parentPath.replaceWithMultiple(newDeclarations);
            }
        }
    };
    export const stateAssignments = (path, { componentCursor, componentHash }) => {
        // STATE
        // FROM state.property = value;
        // TO   setProperty(value);
        if (path.node.type === 'AssignmentExpression') {
            if (path.node.left.type === 'MemberExpression') {
                if (path.node.left.object.type === 'Identifier') {
                    if (path.node.left.object.name === 'state') {
                        // left: path.node.left.property as any
                        if (path.node.left.property.type !== 'Identifier') {
                            return throwError('set state property not yet supported.');
                        }
                        // set that state variable is used!
                        if (componentHash[componentCursor].state[path.node.left.property.name] &&
                            componentHash[componentCursor].state[path.node.left.property.name].hasOwnProperty('used')) {
                            componentHash[componentCursor].state[path.node.left.property.name].used = true;
                        }
                        else {
                            throwError('either component undefined or state set without being defined');
                        }
                        path.replaceWith(t.expressionStatement(t.callExpression(t.identifier(makeSetState(path.node.left.property.name)), [path.node.right])));
                    }
                }
            }
        }
    };
    export const stateFunctions = (path, { componentCursor, componentHash }) => {
        // STATE
        // FROM state.method(value)
        // TO   method(value);
        if (path.node.type === 'CallExpression') {
            if (path.node.callee.type === 'MemberExpression') {
                if (path.node.callee.object.type === 'Identifier') {
                    if (path.node.callee.object.name === 'state') {
                        // left: path.node.callee.property as any
                        if (path.node.callee.property.type !== 'Identifier') {
                            return throwError('set state property not yet supported.');
                        }
                        // set that state variable is used!
                        if (componentHash[componentCursor].state[path.node.callee.property.name] &&
                            componentHash[componentCursor].state[path.node.callee.property.name].hasOwnProperty('used')) {
                            componentHash[componentCursor].state[path.node.callee.property.name].used = true;
                        }
                        else {
                            throwError('either component undefined or state set without being defined');
                        }
                        path.replaceWith(t.callExpression(path.node.callee.property, path.node.arguments));
                    }
                }
            }
        }
    };
    export const defaultProps = (path) => {
        if (path.node.callee.type === 'Identifier' &&
            path.node.callee.name === 'useDefaultProps') {
            const args = path.node.arguments;
            // @ts-ignore
            const Component = args[0];
            // @ts-ignore
            const Props = args[1];
            if (Component.type !== 'Identifier') {
                throwError('useDefaultProps first argument must be an Identifier');
            }
            if (Props.type !== 'ObjectExpression') {
                throwError('useDefaultProps first argument must be an ObjectExpression');
            }
            // replace with Component.defaultProps = Props;
            path.replaceWith(t.expressionStatement(t.assignmentExpression('=', t.memberExpression(t.identifier(Component.name), t.identifier('defaultProps'), false), Props)));
        }
    };
    export const lifecycleHooks = (path) => {
        // FROM
        // onMount(() => {
        //     if (parentRef) {
        //         animateLayout(parentRef);
        //     }
        // });
        // TO 
        // useEffect(() => {
        //     if (parentRef.current) {
        //       animateLayout(parentRef.current);
        //     }
        //   }, []);
        if (path.node.expression.type === 'CallExpression') {
            if (path.node.expression.callee.type === 'Identifier') {
                if (path.node.expression.callee.name === 'onMount') {
                    path.node.expression.callee.name = 'useEffect';
                    path.node.expression.arguments.push(t.arrayExpression([]));
                }
                if (path.node.expression.callee.name === 'onUnmount') {
                    path.node.expression.callee.name = 'useEffect';
                }
                if (path.node.expression.callee.name === 'onUpdate') {
                    path.node.expression.callee.name = 'useEffect';
                }
            }
        }
    };
    export const manageCursor = (path, { componentHash }) => {
        let componentCursor;
        // manage cursor
        if (path.node.init &&
            path.node.init.type === 'CallExpression' &&
            path.node.init.callee &&
            path.node.init.callee.type === 'Identifier' &&
            (path.node.init.callee.name === 'useStore' ||
                path.node.init.callee.name === 'useState' ||
                path.node.init.callee.name === 'useRef')) {
            if (path.parent.type === 'VariableDeclaration') {
                if (path.parentPath.parent.type === 'BlockStatement') {
                    const parent = path.parentPath.parentPath.parent;
                    const parentPath = path.parentPath.parentPath.parentPath;
                    if (parent.type === 'FunctionDeclaration') {
                        const componentName = parent.id.name;
                        componentCursor = componentName;
                        componentHash[componentName] = componentHash[componentName] || {
                            name: componentName,
                            refs: {},
                            state: {}
                        };
                    }
                    else if (parent.type === 'ArrowFunctionExpression') {
                        if (parentPath.parent.type === 'VariableDeclarator') {
                            if (parentPath.parent.id.type === 'Identifier') {
                                const componentName = parentPath.parent.id.name;
                                componentCursor = componentName;
                                componentHash[componentName] = componentHash[componentName] || {
                                    name: componentName,
                                    refs: {},
                                    state: {}
                                };
                            }
                            else {
                                throwError('cannot find ArrowFunctionExpression Component definition');
                            }
                        }
                    }
                    else {
                        throwError('cannot find Component definition');
                    }
                }
            }
        }
        return componentCursor;
    };
};
